"""CLI Command: show."""
from valiant.package import PackageMetadata

from .package_command import PackageCommand, Payload


class ShowCommand(PackageCommand):
    """Describes a package.

    show
        {package : The package name}
        {version : The package version}
        {--r|repository= : The repository to use}
        {--o|out= : the desired output type (json)}
    """

    def prepare_data(self) -> Payload:
        """Gets the package metadata.

        Returns:
            Package metadata

        Raises:
            ValueError: When the package data can't be loaded
        """
        result = self.valiant.get_package_metadata(
            package_name=self.argument("package"),
            package_version=self.argument("version"),
            repository_name=self.option("repository"),
        )

        if not result:
            raise ValueError("Package details could not be loaded.")

        return Payload(metadata=result.package_metadata)

    def to_text(self, data: Payload) -> str:
        """Prepares text representations.

        Args:
            data: A payload that must have the package metadata.

        Returns:
            A nice bit of information to enhance your terminal.

        Raises:
            ValueError: If the payload is missing the package metadata.
        """
        if not data.metadata:
            raise ValueError("No package metadata available.")

        metadata: PackageMetadata = data.metadata
        return f"""\
<info>{metadata.name}:{metadata.version}</info>
<comment>Repository: <link>{metadata.repository_url}</link></comment>
<comment>Summary: {metadata.summary}</comment>"""



"""Base classes for describing a package."""
from abc import abstractmethod
from dataclasses import asdict, dataclass
from datetime import datetime
from typing import Dict, List

from packaging.requirements import Requirement
from valiant.util import Dictionizer

from .classifier import Classifier


class ArtifactMetadata(Dictionizer):
    """Metadata for a package artifact.

    A package version usually has one or more artifacts aligned
    to it. Most releases include a source archive and a built distribution.

    See: https://packaging.python.org/specifications/distribution-formats/

    TODO: Consider the platform requirements for an artifact - e.g. macosx or
    """

    @property
    @abstractmethod
    def comment_text(self) -> str:  # pragma: no cover
        """The comment text provided in the artifact."""

    @property
    @abstractmethod
    def digests(self) -> Dict[str, str]:  # pragma: no cover
        """Provides the hashes (digests) for the artifact.

        For example, refer to
        https://pip.pypa.io/en/stable/reference/pip_install/#hash-checking-mode
        """

    @property
    @abstractmethod
    def sha256_digest(self) -> str:  # pragma: no cover
        """Convenience method to access the SHA256 hash."""

    @property
    @abstractmethod
    def signed(self) -> bool:  # pragma: no cover
        """True if the artifact was signed, False otherwise.

        For further info, check out the --sign option in twine.

        See: https://github.com/pypa/twine#why-should-i-use-this
        """

    @property
    @abstractmethod
    def signature_url(self) -> str:  # pragma: no cover
        """Returns the url of the signing key for the artifact.

        Implementations should provide the URL if supported and for
        artifacts marked `has_sig=True`.

        For example, signatures can be accessed from the central PyPi
        repository by adding `.asc` to the artifact url.
        """

    @property
    @abstractmethod
    def package_type(self) -> str:  # pragma: no cover
        """The packaging type of the artifact.

        Example values:
            - sdist
            - bdist_wheel
        """

    @property
    @abstractmethod
    def python_version(self) -> str:  # pragma: no cover
        """The Python version required for the version.

        Example values:
            - source
            - py2.py3
            - py3

        See: https://www.python.org/dev/peps/pep-0425/
        See: https://www.python.org/dev/peps/pep-0508/#environment-markers
        """

    @property
    @abstractmethod
    def requires_python(self) -> List[str]:  # pragma: no cover
        """A list of Python versions supported by the artifact.

        See: Versioning spec: https://www.python.org/dev/peps/pep-0440/#version-specifiers
        """

    @property
    @abstractmethod
    def size(self) -> int:  # pragma: no cover
        """The size of the artifact in bytes."""

    @property
    @abstractmethod
    def upload_time_iso_8601(self) -> datetime:  # pragma: no cover
        """The ISO 8601-compliant timestamp of the upload.

        See: https://en.wikipedia.org/wiki/ISO_8601
        """

    @property
    @abstractmethod
    def url(self) -> str:  # pragma: no cover
        """The download URL for the artifact."""

    def to_dict(self) -> Dict:  # noqa:DAR102
        return {
            "comment_text": self.comment_text,
            "digests": self.digests,
            "sha256_digest": self.sha256_digest,
            "signed": self.signed,
            "signature_url": self.signature_url,
            "package_type": self.package_type,
            "python_version": self.python_version,
            "requires_python": self.requires_python,
            "size": self.size,
            "upload_time_iso_8601": self.upload_time_iso_8601.isoformat(),
            "url": self.url,
        }


class ArtifactMetadataImpl(ArtifactMetadata):
    """Basic implementation of ArtifactMetadata."""

    def __init__(
        self,
        comment_text: str,
        digests: Dict[str, str],
        sha256_digest: str,
        signed: bool,
        signature_url: str,
        package_type: str,
        python_version: str,
        requires_python: List[str],
        size: int,
        upload_time_iso_8601: datetime,
        url: str,
    ):  # noqa: D107
        self._comment_text = comment_text
        self._digests = digests
        self._sha256_digest = sha256_digest
        self._signed = signed
        self._signature_url = signature_url
        self._package_type = package_type
        self._python_version = python_version
        self._requires_python = requires_python
        self._size = size
        self._upload_time_iso_8601 = upload_time_iso_8601
        self._url = url

    @property
    def comment_text(self) -> str:  # pragma: no cover
        """The comment text provided in the artifact."""
        return self._comment_text  # noqa:DAR201

    @property
    def digests(self) -> Dict[str, str]:  # pragma: no cover
        """Provides the hashes (digests) for the artifact."""
        return self._digests  # noqa:DAR201

    @property
    def sha256_digest(self) -> str:  # pragma: no cover
        """Convenience method to access the SHA256 hash."""
        return self._sha256_digest  # noqa:DAR201

    @property
    def signed(self) -> bool:  # pragma: no cover
        """True if the artifact was signed, False otherwise."""
        return self._signed  # noqa:DAR201

    @property
    def signature_url(self) -> str:  # pragma: no cover
        """Returns the url of the signing key for the artifact."""
        return self._signature_url  # noqa:DAR201

    @property
    def package_type(self) -> str:  # pragma: no cover
        """The packaging type of the artifact."""
        return self._package_type  # noqa:DAR201

    @property
    def python_version(self) -> str:  # pragma: no cover
        """The Python version required for the version."""
        return self._python_version  # noqa:DAR201

    @property
    def requires_python(self) -> List[str]:  # pragma: no cover
        """A list of Python versions supported by the artifact."""
        return self._requires_python  # noqa:DAR201

    @property
    def size(self) -> int:  # pragma: no cover
        """The size of the artifact in bytes."""
        return self._size  # noqa:DAR201

    @property
    def upload_time_iso_8601(self) -> datetime:  # pragma: no cover
        """The ISO 8601-compliant timestamp of the upload."""
        return self._upload_time_iso_8601  # noqa:DAR201

    @property
    def url(self) -> str:  # pragma: no cover
        """The download URL for the artifact."""
        return self._url  # noqa:DAR201


@dataclass
class PackageCoordinates(Dictionizer):
    """An approach to locating a package based on repo, name and version.

    This is based on the Maven approach. It might be useful.

    See: http://maven.apache.org/pom.html
    """

    name: str
    version: str
    repository_url: str

    def to_dict(self) -> Dict:  # noqa: D102
        return asdict(self)

    def __str__(self) -> str:
        """String representation.

        Returns:
            A coordinator described using the classifier format.
        """
        return f"{self.repository_url} :: {self.name} :: {self.version}"


class PackageMetadata(Dictionizer):
    """The metadata for a Python package."""

    def _requirement_to_dict(self, req: Requirement) -> Dict:
        return {
            "name": req.name,
            "url": req.url,
            "specifiers": [
                {
                    "version": s.version,
                    "operator": s.operator,
                    "prereleases": s.prereleases,
                }
                for s in req.specifier
            ],
            "extras": [e for e in req.extras],
            "marker": f"{req.marker}" if req.marker else None,
        }

    def __str__(self) -> str:  # noqa: D105
        return self.to_json()

    def to_dict(self) -> Dict:
        """Extract a Dictionary version of the instance.

        Returns:
            That's right kids, a dictionary.
        """
        return {
            "name": self.name,
            "version": self.version,
            "summary": self.summary,
            "license": self.license,
            "classifiers": self.classifiers,
            "url_project": self.url_project,
            "url_code": self.url_code,
            "url_issue_tracker": self.url_issue_tracker,
            "url_documentation": self.url_documentation,
            "artifacts": [a.to_dict() for a in self.artifacts],
        }

    @property
    def coordinates(self) -> PackageCoordinates:
        """The package coordinates for this metadata."""
        return PackageCoordinates(  # noqa:DAR201
            name=self.name, version=self.version, repository_url=self.repository_url
        )

    @property
    @abstractmethod
    def name(self) -> str:  # pragma: no cover
        """The name of the package."""

    @property
    @abstractmethod
    def summary(self) -> str:  # pragma: no cover
        """A brief summary of the package."""

    @property
    @abstractmethod
    def description(self) -> str:  # pragma: no cover
        """A longer description of the package."""

    @property
    @abstractmethod
    def repository_url(self) -> str:  # pragma: no cover
        """The URL from whence this package info came."""

    @property
    @abstractmethod
    def license(self) -> str:  # pragma: no cover
        """The license assigned to the package.

        Ideally, this will match an SPDX identifier.

        See: https://packaging.python.org/specifications/core-metadata/#license
        See: https://spdx.org/licenses/
        """

    @property
    @abstractmethod
    def version(self) -> str:  # pragma: no cover
        """The package version."""

    @property
    @abstractmethod
    def classifiers(self) -> List[str]:  # pragma: no cover
        """The classifiers listed against the package.

        You can use the Classifier class in this package to parse
        classifiers if needed.
        """

    @property
    @abstractmethod
    def classifiers_parsed(self) -> List[Classifier]:  # pragma: no cover
        """The parsed classifiers listed against the package."""

    @property
    @abstractmethod
    def url_code(self) -> str:  # pragma: no cover
        """The website for the central codebase.

        For example: https://github.com/pomes/valiant
        """

    @property
    @abstractmethod
    def url_documentation(self) -> str:  # pragma: no cover
        """The website providing documentation for the package.

        For example: http://pomes.github.io/valiant
        """

    @property
    @abstractmethod
    def url_project(self) -> str:  # pragma: no cover
        """The project's primary website."""

    @property
    @abstractmethod
    def url_issue_tracker(self) -> str:  # pragma: no cover
        """The issue tracker for the package/project.

        For example: https://github.com/pomes/valiant/issues
        """

    @property
    @abstractmethod
    def requires_python(self) -> List[str]:  # pragma: no cover
        """Lists the Python version specifiers.

        Packaging spec:
        https://packaging.python.org/specifications/core-metadata/#requires-python

        Versioning spec: https://www.python.org/dev/peps/pep-0440/#version-specifiers

        # noqa: DAR202
        Returns:
            The list of requires-python entries
        """

    @property
    @abstractmethod
    def requires_dist(self) -> Dict[str, List[Requirement]]:  # pragma: no cover
        """A distutils-based dependency.

        See:
        https://packaging.python.org/specifications/core-metadata/#requires-dist-multiple-use

        See: PEP 508 -- Dependency specification for Python Software Packages
        https://www.python.org/dev/peps/pep-0508/


        See:
        https://setuptools.readthedocs.io/en/latest/pkg_resources.html#requirement-methods-and-attributes # noqa: B950

        Versioning spec:
            https://www.python.org/dev/peps/pep-0440/#version-specifiers

        # noqa: DAR202
        Returns:
            A dictionary of parsed Requirements with the requirement name as key
        """

    @property
    @abstractmethod
    def artifacts(self) -> List[ArtifactMetadata]:  # pragma: no cover
        """The artifacts related to the package.

        # noqa: DAR202
        Returns:
            A list of artifacts related to the package version.
        """



"""Utility resources for classifiers."""
from dataclasses import asdict, dataclass
from typing import Dict, List

from valiant.util import Dictionizer

SEPARATOR = " :: "

CLASSIFIER_CATEGORY_LICENSE = "License"
CLASSIFIER_CATEGORY_DEVELOPMENT_STATUS = "Development Status"
CLASSIFIER_CATEGORY_ENVIRONMENT = "Environment"
CLASSIFIER_CATEGORY_FRAMEWORK = "Framework"
CLASSIFIER_CATEGORY_INTENDED_AUDIENCE = "Intended Audience"
CLASSIFIER_CATEGORY_NATURAL_LANGUAGE = "Natural Language"
CLASSIFIER_CATEGORY_OPERATING_SYSTEM = "Operating System"
CLASSIFIER_CATEGORY_PROGRAMMING_LANGUAGE = "Programming Language"
CLASSIFIER_CATEGORY_TOPIC = "Topic"
CLASSIFIER_CATEGORY_TYPING = "Typing"


@dataclass
class Classifier(Dictionizer):
    """An individual python classifier.

    See: https://www.python.org/dev/peps/pep-0301/#distutils-trove-classification
    See: https://pypi.org/classifiers/
    """

    original: str
    category: str
    value: str
    subcategories: List[str]

    def to_dict(self) -> Dict:  # noqa:D102
        return asdict(self)

    @staticmethod
    def parse(value: str) -> "Classifier":
        """Parses a classifier string.

        Args:
            value: The classifier string.

        Returns:
            A new Classifier instance based on value.

        Raises:
            ValueError: If value doesn't parse to at least 2 segments.
        """
        components: List[str] = value.split(SEPARATOR)

        subcategories = []

        if len(components) < 2:
            raise ValueError(f"Could not parse classifier: {value}")
        elif len(components) >= 3:
            for sc in components[1:-1]:
                subcategories.append(sc.strip())

        return Classifier(
            original=value,
            category=components[0].strip(),
            value=components[-1].strip(),
            subcategories=subcategories,
        )



"""Base classes for packages and artifacts."""

from .classifier import (
    CLASSIFIER_CATEGORY_DEVELOPMENT_STATUS,
    CLASSIFIER_CATEGORY_ENVIRONMENT,
    CLASSIFIER_CATEGORY_FRAMEWORK,
    CLASSIFIER_CATEGORY_INTENDED_AUDIENCE,
    CLASSIFIER_CATEGORY_LICENSE,
    CLASSIFIER_CATEGORY_NATURAL_LANGUAGE,
    CLASSIFIER_CATEGORY_OPERATING_SYSTEM,
    CLASSIFIER_CATEGORY_PROGRAMMING_LANGUAGE,
    CLASSIFIER_CATEGORY_TOPIC,
    CLASSIFIER_CATEGORY_TYPING,
    Classifier,
)
from .model import (
    ArtifactMetadata,
    ArtifactMetadataImpl,
    PackageCoordinates,
    PackageMetadata,
)



"""Handles the valiant context."""
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Optional

from requests_cache import install_cache
from valiant.log import configure_logging, setup_logging_configuration
from valiant.repositories import RepositoryFactory

from .__about__ import (
    application_copyright_holder,
    application_copyright_year,
    application_description,
    application_homepage,
    application_licence,
    application_name,
    application_tagline,
    application_title,
    application_vendor,
    application_version,
)
from .config import Config
from .package import PackageMetadata
from .reports import ReportFactory, ReportProviderConfiguration, ReportSet
from valiant.repositories import RepositoryConfiguration


@dataclass(frozen=True)
class PythonPackagePayload:
    """A payload that collects package information."""

    repository_base_url: str
    package_name: str
    package_version: str
    package_metadata: PackageMetadata
    reports: Optional[ReportSet] = None

    def clone_with_reports(self, reports: ReportSet) -> "PythonPackagePayload":
        """Creates a new instance with a new report set.

        Args:
            reports: The new report set

        Returns:
            The new instance
        """
        return PythonPackagePayload(
            repository_base_url=self.repository_base_url,
            package_name=self.package_name,
            package_version=self.package_version,
            package_metadata=self.package_metadata,
            reports=reports,
        )


class Valiant:
    """Provides general applications details and instance configuration."""

    def __init__(self, config: Config):
        """Fire up a new valiant.

        Args:
            config: The application configuration
        """
        self._config: Config = config
        self._repo_factory = RepositoryFactory()
        self._report_factory = ReportFactory()

        # Make sure the required directories exist
        self._config.config_dir.mkdir(parents=True, exist_ok=True)
        self._config.cache_dir.mkdir(parents=True, exist_ok=True)

        # Setup logging
        log_config = setup_logging_configuration(
            handlers={
                "default": {
                    "level": "INFO",
                    "formatter": "standard",
                    "class": "logging.handlers.RotatingFileHandler",
                    "filename": Path(self._config.config_dir, "valiant.log"),
                    "maxBytes": 500000,
                    "backupCount": 3,
                }
            },
        )
        configure_logging(log_config)

        # TODO: this is just basic caching for now
        install_cache(
            f"{application_name}-{application_version}-requests-cache",
            backend="sqlite",
            expire_after=86400,
        )

    @property
    def application_version(self) -> str:  # noqa: D102
        return application_version

    @property
    def application_name(self) -> str:  # noqa: D102
        return application_name

    @property
    def application_vendor(self) -> str:  # noqa: D102
        return application_vendor

    @property
    def application_title(self) -> str:  # noqa: D102
        return application_title

    @property
    def application_description(self) -> str:  # noqa: D102
        return application_description

    @property
    def application_tagline(self) -> str:  # noqa: D102
        return application_tagline

    @property
    def application_licence(self) -> str:  # noqa: D102
        return application_licence

    @property
    def application_homepage(self) -> str:  # noqa: D102
        return application_homepage

    @property
    def application_copyright_year(self) -> int:  # noqa: D102
        return application_copyright_year

    @property
    def application_copyright_holder(self) -> str:  # noqa: D102
        return application_copyright_holder

    @property
    def cache_dir(self) -> Optional[Path]:
        """Gets the current instance's cache dir.

        Returns:
            A Path object to the cache directory
        """
        return self._config.cache_dir

    @property
    def config_dir(self) -> Optional[Path]:
        """Gets the current instance's config dir.

        Returns:
            A Path object to the config directory
        """
        return self._config.config_dir

    @property
    def default_repository_name(self) -> str:
        """The default repo name."""
        return self._config.default_repository_name  # noqa: DAR201

    @property
    def repository_configuration(self) -> Dict[str, RepositoryConfiguration]:
        """All the repos.

        Returns:
            The job lot of repo configurations.
        """
        return self._config.repository_configuration

    @property
    def report_configuration(self) -> Dict[str, ReportProviderConfiguration]:
        """The reporting config.

        Returns:
            A dict of the config. Can be empty.
        """
        if not self._config.report_configuration:
            return {}
        return self._config.report_configuration

    def get_package_metadata(
        self,
        package_name: str,
        package_version: str,
        repository_name: Optional[str] = None,
    ) -> PythonPackagePayload:
        """Gets the metadata for the requested package.

        Args:
            package_name: The package name
            package_version: The package version
            repository_name: The name(key) for the repository to use

        Returns:
            A payload with the package metadata.
        """
        if repository_name:
            repo_config = self._config.get_repository_configuration(repository_name)
        else:
            repo_config = self._config.default_repository_configuration

        repo = self._repo_factory.get_repository(repo_config)

        metadata = repo.show(package_name, package_version)

        return PythonPackagePayload(
            repository_base_url=repo_config.base_url,
            package_name=package_name,
            package_version=package_version,
            package_metadata=metadata,
        )

    def get_package_reports(
        self, payload: PythonPackagePayload, reports: List[str] = None
    ) -> PythonPackagePayload:
        """Prepares the reports for the package defined in the payload.

        Args:
            payload: A payload generated by `get_package_metadata`
            reports: A list of the specific reports to run.
                     If no list is provided, all the configured reports are run.

        Returns:
            A new payload instance enhanced with reports.

        Raises:
            ValueError: When no report config is provided.
        """
        if not self._config.report_configuration:
            raise ValueError("No report configuration has been provided.")

        if not reports or len(reports) == 0:
            report_set = self._report_factory.generate_reports(
                package_metadata=payload.package_metadata,
                configuration=self._config.report_configuration,
            )
        else:
            report_configuration: Dict[str, ReportProviderConfiguration] = {}
            for item in reports:
                report_configuration[item] = self._config.report_configuration[item]

            report_set = self._report_factory.generate_reports(
                package_metadata=payload.package_metadata,
                configuration=report_configuration,
            )

        return payload.clone_with_reports(report_set)



"""CLI configuration package."""
from .cli import Cli  # pragma: no cover


def main() -> int:  # pragma: no cover
    """Runs the CLI.

    Returns:
        A status code.
    """
    return Cli().run()



"""Helper assets relating to SPDX (https://spdx.org/about) licenses."""

from dataclasses import asdict, dataclass
from dataclasses import field as dc_field
from datetime import date
from typing import Any, Dict, List, Optional, cast

import requests
from desert import desert
from marshmallow import fields
from marshmallow.exceptions import ValidationError as MarshmallowValidationError
from valiant.util import Dictionizer

"""The canonical SPDX license data in JSON format."""
SPDX_LICENSE_DATA_FILE_URL = (
    "https://raw.githubusercontent.com/spdx/license-list-data/master/json/licenses.json"
)


@dataclass
class SpdxLicense(Dictionizer):
    """An individual license descriptor."""

    name: str
    reference: str
    id: str = cast(str, desert.field(fields.String(data_key="licenseId")))
    is_deprecated: bool = cast(
        bool, desert.field(fields.Boolean(data_key="isDeprecatedLicenseId"))
    )
    details_url: str = cast(str, desert.field(fields.String(data_key="detailsUrl")))
    reference_number: str = cast(
        str, desert.field(fields.String(data_key="referenceNumber"))
    )
    see_also: List[str] = cast(
        List[str], desert.field(fields.List(fields.String, data_key="seeAlso"))
    )
    is_osi_approved: bool = cast(
        bool, desert.field(fields.Boolean(data_key="isOsiApproved"))
    )
    is_fsf_libre: Optional[bool] = dc_field(
        default=None, metadata=desert.metadata(fields.Boolean(data_key="isFsfLibre"))
    )

    def to_dict(self) -> Dict:  # noqa:D102
        return asdict(self)


@dataclass
class _SpdxLicenseListing:
    """Maps against the SPDX JSON format."""

    licenseListVersion: str
    licenses: List[SpdxLicense]
    releaseDate: date


class SpdxLicenses:
    """Represents a set of SPDX licenses."""

    def __init__(self, license_data: _SpdxLicenseListing):
        """New instance.

        You don't really want to call this directly.
        Instead, call one of the `loader` class methods.

        Args:
            license_data: a prepared instance of _SpdxLicenseListing
        """
        self._licenses: Dict[str, SpdxLicense] = {
            license.id: license for license in license_data.licenses
        }
        self._version = license_data.licenseListVersion
        self._release_date: date = license_data.releaseDate

    @property
    def version(self) -> str:
        """The license data version."""
        return self._version  # noqa: DAR201

    @property
    def release_date(self) -> date:
        """The release date of the data file."""
        return self._release_date  # noqa: DAR201

    def get_license(self, id: str) -> Optional[SpdxLicense]:
        """Request a specific license.

        Args:
            id: The requested license ID

        Returns:
            The associated license if available. None otherwise.
        """
        return self._licenses.get(id, None)

    @property
    def licenses(self) -> Dict[str, SpdxLicense]:
        """Returns the full set of licences."""
        return self._licenses  # noqa: DAR201

    @staticmethod
    def loader(data: Dict[str, Any]) -> "SpdxLicenses":
        """Loads the SPDX license data from a JSON string.

        Args:
            data: A dictionary structure in the expected SPDX format.

        Returns:
            An instance of SpdxLicenses if the json_s data could be loaded

        Raises:
            ValueError: If the data cannot be coreectly mapped
        """
        schema = desert.schema(_SpdxLicenseListing, meta={"partial": True})

        try:
            license_data: _SpdxLicenseListing = schema.load(data)
        except (MarshmallowValidationError, TypeError) as e:
            """
            TypeErrors seem to occur in Desert when fields such as
            `licenseId` (mapped to `id`) are missing.
            """
            raise ValueError(f"Could not parse the JSON data: {e}", e)

        return SpdxLicenses(license_data)

    @staticmethod
    def url_loader(url: str = SPDX_LICENSE_DATA_FILE_URL) -> "SpdxLicenses":
        """Loads license data from a URL.

        Args:
            url: The location of the source data.

        Returns:
            An instance of SpdxLicenses if the data could be loaded.

        Raises:
            ValueError: If the URL doesn't work or the data cannot be coreectly mapped
        """
        r = requests.get(url)
        if r.status_code != requests.codes.ok:
            raise ValueError(f"No result for {url}. Status code: {r.status_code}")

        return SpdxLicenses.loader(r.json())

    @staticmethod
    def builtin_loader() -> "SpdxLicenses":
        """Loads the license data provided by SPDX.

        A pre-built pickle file is loaded from this package. The file is
        generated by the project team using the script in `__main__.py`.

        Returns:
            The set of SPDX licences if it can be restored.

        Raises:
            ValueError: When the data cannot be loaded.
        """
        from importlib import resources as pkg_resources
        import pickle  # noqa: S403

        DATA_PACKAGE = "valiant.reports.spdx"
        DATA_FILE = "spdx-licenses.pickle"

        if not pkg_resources.is_resource(DATA_PACKAGE, DATA_FILE):  # pragma: no cover
            raise ValueError("Failed to access the data in the package.")

        with pkg_resources.open_binary(DATA_PACKAGE, DATA_FILE) as p:
            data = pickle.load(p)  # noqa: S301

        if type(data) is SpdxLicenses:
            return data

        """This is a saftey net in case the pickle file is dodgy."""
        raise ValueError(
            "The loaded datafile did not match the expected structure."
        )  # pragma: no cover



"""The primary entrypoint into the valiant project."""
import os

from .valiant import Valiant


_ROOT = os.path.dirname(os.path.realpath(__file__))


class Factory:
    """Generate new Valiant instances."""

    from typing import Optional
    from .config import Config

    def create_valiant(self, config: Optional[Config] = None) -> Valiant:
        """Generates a valiant instance.

        TBD - this factory is rather underdone atm.

        Args:
            config: A Valiant configuration instance.

        Returns:
            A freshly built valiant
        """
        from .config.config import _load_default_config

        if not config:
            conf = _load_default_config()
        else:
            conf = config

        return Valiant(config=conf)



"""Utility cli."""
import argparse
import pickle  # noqa: S403

from . import SPDX_LICENSE_DATA_FILE_URL, SpdxLicenses


"""Builds a pickle file from the SPDX License json data.

This function generates the `spdx-licenses.pickle` file found in this package.

Example:
    python -m valiant.reports.spdx valiant/reports/spdx/spdx-licenses.pickle

Returns:
    The path of the pickle file
"""


parser = argparse.ArgumentParser(
    description="Pickle the SPDX license data.",
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
)
parser.add_argument(
    "--url",
    type=str,
    default=SPDX_LICENSE_DATA_FILE_URL,
    help="The url for the SPDX License JSON file",
)

parser.add_argument("output_file", type=str, help="The file path for the pickle file")

args = parser.parse_args()

licenses: SpdxLicenses = SpdxLicenses.url_loader(args.url)
with open(args.output_file, "wb") as f:
    pickle.dump(licenses, f)



"""Models for reports and report providers."""
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import Dict, Iterable, List, Optional, Tuple

from valiant.util import Dictionizer
from valiant.package import PackageCoordinates


class NoValue(Enum):
    """As per https://docs.python.org/3/library/enum.html#omitting-values."""

    def __repr__(self):  # noqa:D105
        return "<%s.%s>" % (self.__class__.__name__, self.name)


class FindingLevel(NoValue):
    """A priority level."""

    PRIORITY = "priority"
    WARNING = "warning"
    INFO = "info"


class FindingCategory(NoValue):
    """A useful set of finding categories.

    This is just a helper and you should use the value
    of the member/instance (FindingCategory.SECURITY.value).
    """

    SECURITY = "security"
    LICENSE = "license"
    PROJECT = "project"


@dataclass(frozen=True)
class Finding(Dictionizer):
    """A single finding note."""

    coordinates: PackageCoordinates

    """ An identifier for the finding.

    Please use the provider name (in uppercase) followed by a 3-digit identifier
    such as 001 and 078. Examples include:
        - BASIC001
        - SPDX021
        - SAFETY004
    """
    id: str

    """An appropriate priority for this finding."""
    level: FindingLevel

    """An open category.

    Where possible, use FindingCategory values (e.g. FindingCategory.SECURITY.value)."""
    category: str

    """A brief (<100 characters) title for the finding."""
    title: Optional[str]

    """A longer description of the finding."""
    message: str

    """Allows report providers to optionally add further data."""
    data: Optional[Dictionizer]

    """A link for further details."""
    url: Optional[str]

    def to_dict(self) -> Dict:  # noqa:D102
        data = None

        if self.data:
            data = self.data.to_dict()

        return {
            "id": self.id,
            "coordinates": self.coordinates.to_dict(),
            "level": self.level.value,
            "category": self.category,
            "title": self.title,
            "message": self.message,
            "data": data,
            "url": self.url,
        }


@dataclass(frozen=True)
class ReportProviderDetails(Dictionizer):
    """Describes the report provider."""

    name: str
    vendor: Optional[str]
    display_name: str

    """The version of the report provider."""
    version: str

    """A link for further details."""
    url: Optional[str]

    def to_dict(self) -> Dict:  # noqa:D102
        return {
            "name": self.name,
            "vendor": self.vendor,
            "display_name": self.display_name,
            "version": self.version,
            "url": self.url,
        }


@dataclass(frozen=True)
class ReportProviderConfiguration:
    """A very generic class to allow for flexible configuration."""

    items: Optional[Dict[str, str]] = None

    def get(self, key: str, default: str = None) -> Optional[str]:
        """Retrieve a configuration item.

        Args:
            key: The item key
            default: a value to return if the key doesn't exist

        Returns:
            The value of the config item assigned to the key.
            default value otherwise
        """
        if self.items:
            return self.items.get(key, default)

        return default


class Report(Dictionizer):
    """An individual report from a report provider."""

    def __init__(self, provider_details: ReportProviderDetails):
        """Constructor.

        Args:
            provider_details: The provider's details.
        """
        self._findings: Dict[FindingLevel, List[Finding]] = {}

        self._created: datetime = datetime.now()
        self._provider: ReportProviderDetails = provider_details

    def add_finding(self, finding: Finding) -> None:
        """Add a finding to the list."""  # noqa:DAR101
        if finding.level not in self._findings:
            self._findings[finding.level] = [finding]
        else:
            self._findings[finding.level].append(finding)

    def add_findings(self, findings: List[Finding]) -> None:
        """Add a finding to the list."""  # noqa:DAR101
        for finding in findings:
            self.add_finding(finding)

    @property
    def created(self) -> datetime:  # noqa:D102
        return self._created

    @property
    def provider_details(self) -> ReportProviderDetails:  # noqa:D102
        return self._provider

    @property
    def findings(self) -> Dict[FindingLevel, List[Finding]]:  # noqa:D102
        return self._findings

    @property
    def all_findings(self) -> List[Finding]:  # noqa:D102
        from itertools import chain

        return list(chain.from_iterable(self._findings.values()))

    def to_dict(self) -> Dict:  # noqa:D102
        return {
            "created": self.created.isoformat(),
            "provider": self.provider_details.to_dict(),
            "findings": [finding.to_dict() for finding in self.all_findings],
        }


class ReportSet(Dictionizer):
    """Manages a set of reports from various providers."""

    def __init__(self):
        """Constructor."""
        self._reports: Dict[str, Report] = {}

    def add_report(self, report: Report) -> None:
        """Add a single report to the set.

        Args:
            report: The report to add (no, really)

        Raises:
            ValueError: when the `report.provider_details.name` has
                        already filed a report
        """
        name = report.provider_details.name
        if name in self._reports:
            raise ValueError(f"The provider {name} already has a report in the set.")

        self._reports[name] = report

    def items(self) -> Iterable[Tuple[str, Report]]:
        """Lets you iterate over the reports in the set.

        Returns:
            As for a dictionary
        """
        return self._reports.items()

    @property
    def all_findings(self) -> List[Finding]:
        """Get all findings across the reports.

        Returns:
            A list of all findings from across all reports in the set.
        """
        from itertools import chain

        return list(
            chain.from_iterable(
                [report.all_findings for report in self._reports.values()]
            )
        )

    def to_dict(self) -> Dict:  # noqa:D102
        return {name: report.to_dict() for name, report in self._reports.items()}



"""A centralised set of config information."""
from .__version__ import __version__


application_name = "valiant"
application_title = "Valiant"
application_version = __version__
application_vendor = "Pomes"
application_tagline = "Dependency Investigations Unit"
application_description = "Valiant helps you investigate dependencies"
application_licence = "MIT"
application_copyright_year = 2020
application_copyright_holder = "Duncan Dickinson"
application_homepage = "https://github.com/pomes/valiant"
application_config_file = f"{application_name}.toml"
application_envvar_prefix = "VALIANT_"
pyproject_config_file = "pyproject.toml"
user_config_file = "config.toml"



"""CLI Command: report."""
from .package_command import PackageCommand, Payload
from .report_helper import (
    create_short_report,
    create_metadata_report,
    create_finding_report,
)


class ReportCommand(PackageCommand):
    """Prepares a report on the package.

    report
        {package : The package name}
        {version : The package version}
        {reports? : One or more reports to run (optional - default is all reports)}
        {--r|repository= : The repository to use (not used)}
        {--s|short : Single table output}
        {--o|out= : the desired output type (json)}
    """

    def prepare_data(self) -> Payload:
        """Gets the package metadata.

        Returns:
            Package metadata

        Raises:
            ValueError: When the package data can't be loaded
        """
        package_metadata = self.valiant.get_package_metadata(
            package_name=self.argument("package"),
            package_version=self.argument("version"),
            repository_name=self.option("repository"),
        )

        if self.argument("reports"):
            report_list = self.argument("reports").split(",")
        else:
            report_list = None

        if not package_metadata:
            raise ValueError("Package details could not be loaded.")

        payload = self.valiant.get_package_reports(
            package_metadata, reports=report_list
        )

        return Payload(metadata=payload.package_metadata, reports=payload.reports)

    def to_text(self, payload: Payload) -> str:
        """Prepares text representations.

        Args:
            payload: A payload with the package metadata and reports

        Returns:
            Text just for you.

        Raises:
            ValueError: When the package metadata or reports are missing

        """
        if not payload.metadata:
            raise ValueError("No package metadata available.")

        if not payload.reports:
            raise ValueError("The report set has not been provided.")

        if self.option("short"):
            output = create_short_report(payload.reports.all_findings)
        else:
            output = create_metadata_report(payload.metadata)

            for _, report in payload.reports.items():
                output += create_finding_report(payload.metadata, report)

        return output



"""Configuration for Valiant."""
from copy import deepcopy
from pathlib import Path
from typing import Dict, List, Optional

from valiant.repositories import RepositoryConfiguration
from valiant.repositories.pypi import PyPiRepository
from valiant.reports import ReportProviderConfiguration


class Config:
    """Configuration object."""

    def __init__(
        self,
        repository_configurations: List[RepositoryConfiguration],
        config_dir: str,
        cache_dir: str,
        default_repository: Optional[str] = None,
        report_configuration: Optional[Dict[str, ReportProviderConfiguration]] = None,
    ):
        """Constructor.

        Args:
            repository_configurations: A list of repository configurations.
            cache_dir: Used by Valiant to cache assets.
            config_dir: Contains the configuration files.
            default_repository: The repository configuration to use by default
                                Only needed if there are > 1 repositories
            report_configuration: The report types to run by default

        Raises:
            ValueError: When the `repositories` list has items with the same name.
        """
        if len(repository_configurations) == 0:
            raise ValueError("No repositories were provided.")
        elif not default_repository and len(repository_configurations) > 1:
            raise ValueError("A default repository is required.")

        self._cache_dir = Path(cache_dir)
        self._config_dir = Path(config_dir)

        self._report_configuration: Optional[
            Dict[str, ReportProviderConfiguration]
        ] = None
        if report_configuration:
            self._report_configuration = deepcopy(report_configuration)

        self._repositories: Dict[str, RepositoryConfiguration] = {}

        for repo in repository_configurations:
            if repo.name in self._repositories:
                raise ValueError(f"Found repositories with the same name: {repo.name}")
            else:
                self._repositories[repo.name] = deepcopy(repo)

        if len(self._repositories) == 1 and not default_repository:
            self._default_repository = next(iter(self._repositories.keys()))
        elif default_repository:
            self._default_repository = default_repository
        else:  # pragma: no cover
            raise ValueError("Could not reconcile repository configuration.")

        if self._default_repository not in self._repositories:
            raise ValueError(
                f"The default repository ({self._default_repository})"
                " was not provided in the list of repositories."
            )

    @property
    def cache_dir(self) -> Path:
        """Path to the cache."""
        return self._cache_dir  # noqa: DAR201

    @property
    def config_dir(self) -> Path:
        """Path to the config dir."""
        return self._config_dir  # noqa: DAR201

    @property
    def report_configuration(self) -> Optional[Dict[str, ReportProviderConfiguration]]:
        """The reporting config."""
        return self._report_configuration  # noqa: DAR201

    @property
    def default_repository_name(self) -> str:
        """The default repo name."""
        return self._default_repository  # noqa: DAR201

    @property
    def default_repository_configuration(self) -> RepositoryConfiguration:
        """The default repo.

        Returns:
            If only one repository is listed then it is returned.
            Otherwise, the repository designated as the default_repository is returned.
        """
        return self.get_repository_configuration(self.default_repository_name)

    @property
    def repository_configuration(self) -> Dict[str, RepositoryConfiguration]:
        """All the repos.

        Returns:
            The job lot of repo configurations.
        """
        return self._repositories

    @property
    def repository_names(self) -> List[str]:
        """Returns a list of the known repositories.

        Use the names to call `get_repository`.

        Returns:
            A list of repository names.
        """
        return list(self._repositories.keys())

    def get_repository_configuration(self, key: str) -> RepositoryConfiguration:
        """Lookup a repository.

        Args:
            key: The name of the repository

        Returns:
            The repository configuration

        Raises:
            KeyError: If the name can't be found.
        """
        try:
            repo = self._repositories[key]
        except KeyError as ke:
            raise KeyError(f"Repository with name {key} could not be found.") from ke

        return repo


def _load_default_config() -> Config:
    """Returns an instance with the defaults.

    Returns:
        A default Config instance.
    """
    from appdirs import AppDirs

    from valiant.__about__ import (
        application_name,
        application_vendor,
        application_version,
    )

    dirs = AppDirs(
        appname=application_name,
        appauthor=application_vendor,
        version=application_version,
    )
    return Config(
        repository_configurations=[PyPiRepository.get_pypi_config()],
        config_dir=dirs.user_config_dir,
        cache_dir=dirs.user_cache_dir,
        report_configuration={
            "basic": ReportProviderConfiguration(),
            "spdx": ReportProviderConfiguration(),
            "safety": ReportProviderConfiguration(),
        },
    )



"""The much adored central Python repo."""

import sys

from pathlib import Path
from typing import Any, Dict, List

import requests

from valiant.log import get_logger
from valiant.repositories import (
    BaseRepository,
    PackageNotFoundException,
    RepositoryConfiguration,
    RepositoryException,
)
from .model import PyPiPackageMetadata

log = get_logger()


class PyPiRepository(BaseRepository):
    """The central Python repository.

    This class may also work on servers also running the Warehouse system.

    See: https://warehouse.readthedocs.io/api-reference/json/
    """

    @classmethod
    def list_supported_repository_types(cls) -> List[str]:
        """Lists the repository types support by this implementation."""
        return ["warehouse", "pypi"]  # noqa: DAR201

    def _load_package_manifest(self, name: str, version: str) -> Dict[Any, Any]:
        """Downloads the JSON metadata from the repository.

        Args:
            name: The package name.
            version: The package version.

        Returns:
            The JSON-based metadata (as a string).

        Raises:
            PackageNotFoundException: When the URL doesn't work.
            RepositoryException: If the JSON data is empty(None)
        """
        url = f"{self.repository_configuration.get_access_url()}/{name}/{version}/json"

        r = requests.get(url)
        if r.status_code != requests.codes.ok:
            log.error(
                "Package not found",
                package_name=name,
                package_version=version,
                repository_url=self.repository_configuration.base_url,
                status_code=r.status_code,
            )
            raise PackageNotFoundException(f"No result for {url}")

        log.info(
            "Package found",
            package_name=name,
            package_version=version,
            repository_url=self.repository_configuration.base_url,
            cache_used=r.from_cache,  # type: ignore
        )

        data = r.json()

        if data is None:
            raise RepositoryException("The JSON response was empty.")

        return data

    def show(self, name: str, version: str) -> PyPiPackageMetadata:
        """Provides details for a specific package version.

        Args:
            name: The package name
            version: The package version

        Returns:
            A package instance if it can be located. None otherwise.

        Raises:
            PackageNotFoundException: When the package cannot be found
            ValidationError: The package metadata could not be parsed. # noqa: DAR402
        """
        try:
            data = self._load_package_manifest(name, version)
        except (PackageNotFoundException, RepositoryException) as e:
            raise PackageNotFoundException(f"Failed to access package metadata: {e}")

        return PyPiPackageMetadata(
            repository_url=self.repository_configuration.base_url, package_data=data
        )

    def download(self, name: str, version: str) -> Path:
        """Provides details for a specific package version.

        Args:
            name: The package name
            version: The package version

        # noqa: DAR202
        Returns:
            The path to the download if the artifact can be located.

        Raises:
            NotImplementedError: Because this is an abstract implementation.
        """
        raise NotImplementedError

    @staticmethod
    def get_pypi_config() -> RepositoryConfiguration:
        """Helper for the central repo.

        Returns:
            A config for the PyPi repo
        """
        return RepositoryConfiguration(
            name="pypi", base_url="https://pypi.org/pypi", repository_type="warehouse"
        )


if __name__ == "__main__":
    """Just a small cli.

    Sample call:
        poetry run python -m valiant.repositories.pypi.pypi flask 1.1.1
    """

    if len(sys.argv) == 3:
        repo = PyPiRepository(PyPiRepository.get_pypi_config())
        pkg = repo.show(name=sys.argv[1], version=sys.argv[2])

        print(f"{pkg}")
    else:
        print("Usage: pypi <package> <version>")



"""Classes related to working with Python package repositories."""

from .config import RepositoryConfiguration
from .exceptions import PackageNotFoundException, RepositoryException, ValidationError
from .repository import BaseRepository


class RepositoryFactory:
    """Helps construct a repository instance based on the configuration."""

    def __init__(self):
        """Constructor."""
        from typing import Dict

        self._cache: Dict[str, BaseRepository] = {}

    def _instantiate_handler(self, conf: RepositoryConfiguration) -> BaseRepository:
        from .pypi import PyPiRepository

        if conf.repository_type in PyPiRepository.list_supported_repository_types():
            return PyPiRepository(conf)

        raise ValueError(
            f"Unable to handle repositories of type {conf.repository_type}"
        )

    def _check_cache(self, conf: RepositoryConfiguration) -> BaseRepository:
        if conf.name in self._cache:
            if self._cache[conf.name].repository_configuration != conf:
                raise ValueError(
                    "Cache clash"
                    " - repository configuration uses the same name as an existing cache entry"
                )
        else:
            self._cache[conf.name] = self._instantiate_handler(conf)

        return self._cache[conf.name]

    def get_repository(
        self, repository_configuration: RepositoryConfiguration,
    ) -> BaseRepository:
        """Factory method.

        Args:
            repository_configuration: The repository config

        Returns:
            A repository instance that handles the type set in the config
        """
        return self._check_cache(repository_configuration)

    def reset_cache(self) -> None:
        """Clears the cache."""
        self._cache = {}



"""CLI Command: audit."""
from pathlib import Path
from typing import List, NamedTuple, Optional

from valiant.reports import Finding, ReportSet

from .base_command import BaseCommand
from .package_command import Payload
from .report_helper import (
    create_metadata_report,
    create_finding_report,
    create_short_report,
)


class AuditCommand(BaseCommand):
    """Prepares an audit on a set of requirements (not yet implemented).

    audit
        {requirements-file : The file containing a requirements list}
        {reports? : One or more reports to run (optional - default is all reports)}
        {--s|short : Single table output}
        {--o|out= : the desired output type (json)}

    The audit command expects a very basic requirements file with one line per requirement
    and each requirement pinned to a specific version (e.g. texttable==1.6.2)

    Poetry users can generate this file with:
        poetry export --without-hashes --format requirements.txt --output requirements.txt

    """

    class _RequirementsEntry(NamedTuple):
        package: str
        version: str

    def handle(self) -> Optional[int]:  # noqa: D102
        requirements = Path(self.argument("requirements-file"))
        report_list: Optional[List[str]] = None
        format = self.option("out")
        package_list: List[AuditCommand._RequirementsEntry] = []
        payloads: List[Payload] = []

        if self.argument("reports"):
            report_list = self.argument("reports").split(",")

        try:
            if not requirements.is_file():
                raise ValueError(
                    f"The requirements file ({requirements}) doesn't exist."
                )

            with open(requirements, "r") as reqs:
                for line in reqs:
                    package, version = line.strip().split("==")
                    package_list.append(
                        AuditCommand._RequirementsEntry(package, version)
                    )

            for req in package_list:
                package_metadata = self.valiant.get_package_metadata(
                    package_name=req.package, package_version=req.version,
                )
                reports = self.valiant.get_package_reports(
                    package_metadata, reports=report_list
                )
                payloads.append(
                    Payload(metadata=reports.package_metadata, reports=reports.reports)
                )

            if format == "json":
                self.line(self.to_json(payloads))
            else:
                self.line(self.to_text(payloads))

        except Exception as e:
            self.output_error(e, format=self.option("out"))
            return 1

        return 0

    def compile_all_findings(self, payloads: List[Payload]) -> List[Finding]:
        """Compiles all findings across all reports.

        Args:
            payloads: A list of payloads with reportsets

        Returns:
            A list containing all findings
        """
        from itertools import chain

        all_report_sets: List[ReportSet] = [
            payload.reports for payload in payloads if payload.reports
        ]

        return list(chain.from_iterable([rs.all_findings for rs in all_report_sets]))

    def prepare_payload_report(self, payload: Payload) -> str:
        """Draws up yer reports.

        Args:
            payload: A single package payload

        Returns:
            A nice report for you and your pals

        Raises:
            ValueError: if you forget to load the payload with the metadata and reports data
        """
        if not payload.metadata:
            raise ValueError("No package metadata available.")

        if not payload.reports:
            raise ValueError("The report set has not been provided.")

        output = create_metadata_report(payload.metadata)

        for _, report in payload.reports.items():
            output += create_finding_report(payload.metadata, report)

        return output

    def to_text(self, payloads: List[Payload]) -> str:
        """Prepares text representations.

        Args:
            payloads: A list of payloads with the package metadata and reports

        Returns:
            Text just for you.
        """
        if self.option("short"):
            output = create_short_report(self.compile_all_findings(payloads))
        else:
            outlist = []
            for payload in payloads:
                outlist.append(self.prepare_payload_report(payload))

            output = "\n".join(outlist)

        return output

    def to_json(self, payloads: List[Payload]) -> str:
        """Converts data to json."""  # noqa:DAR101,DAR201
        import json

        if self.option("short"):
            return json.dumps(
                [f.to_dict() for f in self.compile_all_findings(payloads)]
            )

        return json.dumps([p.to_dict() for p in payloads])



"""Logging for the Valiant system.

This is an initial logging setup and may be underwhelming.
"""
from copy import deepcopy
import logging
from logging.config import dictConfig
import sys
import structlog
from typing import Dict
from structlog.stdlib import LoggerFactory


"""A basic default that spits out to standard out."""
DEFAULT_CONFIG = {
    "version": 1,
    "disable_existing_loggers": True,
    "formatters": {"standard": {"format": "%(message)s"}},
    "handlers": {
        "default": {
            "level": "INFO",
            "formatter": "standard",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stdout",  # Default is stderr
        },
    },
    "loggers": {
        "": {  # root logger
            "handlers": ["default"],
            "level": "INFO",
            "propagate": False,
        }
    },
}


def setup_logging_configuration(
    disable_existing_loggers: bool = True,
    handlers: Dict = None,
    formatters: Dict = None,
    loggers: Dict = None,
) -> Dict:
    """Convenience function to structure the logging config.

    This method essentially layers over the top of the default config.

    See: https://docs.python.org/3/library/logging.config.html#logging.config.dictConfig
    See: https://docs.python.org/3/library/logging.config.html#logging-config-dictschema

    Args:
        disable_existing_loggers: as per the configuration dictionary schema
        handlers: as per the configuration dictionary schema
        formatters: as per the configuration dictionary schema
        loggers: as per the configuration dictionary schema

    Returns:
        A ready-to-go logging config.
    """
    config = deepcopy(DEFAULT_CONFIG)
    config["disable_existing_loggers"] = disable_existing_loggers
    if handlers:
        config["handlers"] = handlers
    if formatters:
        config["formatters"] = formatters
    if loggers:
        config["loggers"] = loggers
    return config


def configure_logging(config: Dict = DEFAULT_CONFIG) -> None:
    """Configures logging for the application.

    This is called by Valiant.__init__

    See: https://docs.python.org/3/library/logging.config.html#logging.config.dictConfig

    Args:
        config: A dictionary acceptable to logging.config.dictConfig
    """
    dictConfig(deepcopy(config))

    # See: http://www.structlog.org/en/stable/standard-library.html#rendering-within-structlog
    structlog.configure_once(
        processors=[
            structlog.stdlib.filter_by_level,
            structlog.stdlib.add_logger_name,
            structlog.stdlib.add_log_level,
            structlog.stdlib.PositionalArgumentsFormatter(),
            structlog.processors.TimeStamper(fmt="iso"),
            structlog.processors.StackInfoRenderer(),
            structlog.processors.format_exc_info,
            structlog.processors.UnicodeDecoder(),
            structlog.processors.JSONRenderer(),
        ],
        context_class=dict,
        logger_factory=structlog.stdlib.LoggerFactory(),
        wrapper_class=structlog.stdlib.BoundLogger,
        cache_logger_on_first_use=True,
    )


def get_logger():  # noqa:ANN201
    """Get a logger for your needs.

    Returns:
        An object that allows you to call standard logging methods
        but also pass in arbitrary keyword params (kwargs) that are
        put out to the log in a handy parseable form

    Example usage:

        from valiant.log import get_logger
        log = get_logger()

        log.info(
            f"SPDX reporter located {len(report.findings)} findings",
            package_name=package_metadata.name,
            package_version=package_metadata.version,
            repository_url=package_metadata.repository_url,
        )
    """
    return structlog.get_logger()



"""Report generation functionality."""
from valiant.package import PackageMetadata

from .model import (
    Finding,
    FindingCategory,
    FindingLevel,
    NoValue,
    Report,
    ReportProviderConfiguration,
    ReportProviderDetails,
    ReportSet,
)
from .report import BaseReportProvider
from .basic import BasicReportProvider
from .spdx import SpdxLicenseReportProvider
from .safety import SafetyReportProvider


class ReportFactory:
    """Factory for generating Valiant reports."""

    from typing import Dict

    def generate_reports(
        self,
        package_metadata: PackageMetadata,
        configuration: Dict[str, ReportProviderConfiguration],
    ) -> ReportSet:
        """Generate a requested report.

        The report configuration is used to determine the report provider.

        Args:
            package_metadata: The python package metadata.
            configuration: The report providers to use.

        Returns:
            A ReportSet instance with reports for all requested providers.

        Raises:
            ValueError: When a requested provider isn't available.
        """
        reports = ReportSet()
        for provider, config in configuration.items():
            if provider == "basic":
                reports.add_report(
                    BasicReportProvider(config).generate_report(package_metadata)
                )
            elif provider == "safety":
                reports.add_report(
                    SafetyReportProvider(config).generate_report(package_metadata)
                )
            elif provider == "spdx":
                spdx = SpdxLicenseReportProvider(config)
                report = spdx.generate_report(package_metadata)
                reports.add_report(report)
            else:
                raise ValueError(
                    f"No implementation available for provider '{provider}'"
                )

        return reports



"""Trait for dictionary and json translation."""
import json

from abc import ABC, abstractmethod
from typing import Dict


class Dictionizer(ABC):
    """Implementers provide a dict and json rendition."""

    @abstractmethod
    def to_dict(self) -> Dict:
        """Convert the instance to a dictionary format."""
        raise NotImplementedError()  # noqa:DAR401

    def to_json(self) -> str:
        """Convert the instance to JSON format."""  # noqa:DAR201
        return json.dumps(self.to_dict())



"""CLI Command: config."""
import json
from typing import Optional
from .base_command import BaseCommand


class ConfigCommand(BaseCommand):
    """Application configuration.

    config
        {--o|out= : the desired output type (json)}
    """

    def handle(self) -> Optional[int]:  # noqa: D102
        format = self.option("out")
        if format == "json":
            self.line(self.to_json())
        else:
            self.line(self.to_text())
        return 0

    def to_json(self) -> str:  # noqa: D102
        return json.dumps(
            {
                "cache_dir": str(self.valiant.cache_dir),
                "config_dir": str(self.valiant.config_dir),
                "default_repository_name": self.valiant.default_repository_name,
                "repositories": {
                    k: v.to_dict()
                    for k, v in self.valiant.repository_configuration.items()
                },
                "reports": {
                    k: v.items for k, v in self.valiant.report_configuration.items()
                },
            }
        )

    def to_text(self) -> str:  # noqa: D102
        repo_config = ",".join(self.valiant.repository_configuration.keys())
        report_config = ",".join(self.valiant.report_configuration.keys())
        return f"""\
<info>cache_dir</info>: <comment>{self.valiant.cache_dir}</comment>
<info>config_dir</info>: <comment>{self.valiant.config_dir}</comment>
<info>default_repository_name</info>: <comment>{self.valiant.default_repository_name}</comment>
<info>repositories</info>: <comment>{repo_config}</comment>
<info>reports</info>: <comment>{report_config}</comment>"""



"""Helper funcs for generating reports."""
from typing import List

from texttable import Texttable
from valiant.package import PackageMetadata
from valiant.reports import Finding, Report


def create_metadata_report(metadata: PackageMetadata) -> str:
    """Reports on the package metadata.

    # noqa:DAR101
    # noqa:DAR201
    """
    base_package_table = Texttable()

    base_package_table.add_rows(
        [
            ["Item", "Value(s)"],
            ["Package", metadata.name],
            ["Version", metadata.version],
            ["Repository", metadata.repository_url],
            ["License", metadata.license],
            ["Summary", metadata.summary],
            [
                "Resources",
                "\n".join(
                    [
                        f"Project: {metadata.url_project}",
                        f"Code: {metadata.url_code}",
                        f"Documentation: {metadata.url_documentation}",
                        f"Issue tracker:: {metadata.url_issue_tracker}",
                    ]
                ),
            ],
            ["Classifiers", "\n".join(metadata.classifiers)],
            [
                "Artifacts",
                "\n---\n".join(
                    [f"{a.package_type}: {a.url}" for a in metadata.artifacts]
                ),
            ],
        ]
    )

    return f"<h1>Package metadata</h1>\n{base_package_table.draw()}\n\n"


def create_finding_report(metadata: PackageMetadata, report: Report) -> str:
    """Reports on the findings.

    # noqa:DAR101
    # noqa:DAR201
    """
    if len(report.findings.keys()) == 0:
        findings_report = "<comment>No findings</comment>"
    else:
        findings_table = Texttable()
        findings_table.set_cols_dtype(["t", "t", "t", "t", "t"])
        findings_table.set_cols_align(["c", "c", "c", "l", "l"])
        findings_table.header(["Priority", "ID", "Category", "Title", "Message"])
        for finding in report.all_findings:
            findings_table.add_row(
                [
                    finding.level.value,
                    finding.id,
                    finding.category,
                    finding.title,
                    finding.message,
                ]
            )

        findings_report = findings_table.draw()

    report_title = (
        f"Report: {report.provider_details.display_name} [{metadata.coordinates}]"
    )

    return f"<h1>{report_title}</h1>\n{findings_report}\n\n"


def create_short_report(findings: List[Finding]) -> str:
    """Produces a table of findings.

    Args:
        findings: A list of findings

    Returns:
        A nice table for you
    """
    report_table = Texttable()
    report_table.header(
        ["Package Coordinates", "ID", "Level", "Category", "Title", "Message"]
    )

    for finding in findings:
        report_table.add_row(
            [
                finding.coordinates,
                finding.id,
                finding.level.value,
                finding.category,
                finding.title,
                finding.message,
            ]
        )

    return report_table.draw()



"""A helpful base to get started from."""
import json
import traceback

from cleo import Command as CleoCommand
from clikit.api.io.flags import VERBOSE

from valiant import Valiant


class BaseCommand(CleoCommand):
    """Common methods for most commands."""

    @property
    def valiant(self) -> Valiant:
        """Access the Valiant application object."""
        return self.application.valiant  # noqa: DAR201

    def output_error(self, e: Exception, format: str = None) -> None:
        """Outputs the error to the required format."""  # noqa:DAR101
        if format == "json":
            data = {"error": {"type": e.__class__.__name__, "message": str(e)}}
            if self.io.is_verbose():
                data["error"]["trace"] = traceback.format_exc()
            self.line_error(json.dumps(data))
        else:
            output = f"<error>{e.__class__.__name__}: {e}\n"
            self.line_error(output)
            if self.io.is_verbose():
                self.line_error(
                    f"<debug>{traceback.format_exc()}</debug>", verbosity=VERBOSE
                )



"""Package exceptions."""


class RepositoryException(Exception):  # noqa: D101

    pass


class PackageNotFoundException(RepositoryException):  # noqa: D101

    pass


class ValidationError(RepositoryException):  # noqa: D101

    pass



"""A basic package reporter."""
from enum import Enum

from valiant.log import get_logger
from valiant.package import PackageCoordinates, PackageMetadata
from valiant.reports import (
    BaseReportProvider,
    Finding,
    FindingCategory,
    FindingLevel,
    Report,
    ReportProviderDetails,
)
from valiant.util import Dictionizer


log = get_logger()


class BasicId(Enum):
    """Setup for findings."""

    LICENCE_NOT_FOUND = (
        "BASIC001",
        FindingLevel.PRIORITY,
        "No license found",
        "The project's package needs to declare a license.",
    )
    PROJECT_NOT_FOUND = (
        "BASIC002",
        FindingLevel.WARNING,
        "No link to project site",
        "The project doesn't provide a link to its project site.",
    )
    CODEBASE_NOT_FOUND = (
        "BASIC003",
        FindingLevel.WARNING,
        "No link to codebase",
        "The project doesn't provide a link to its codebase.",
    )
    ARTIFACT_NOT_SIGNED = (
        "BASIC004",
        FindingLevel.WARNING,
        "An artifact has not been signed",
        "One of the artifacts has not been signed.",
    )
    NON_PRODUCTION_RELEASE = (
        "BASIC005",
        FindingLevel.WARNING,
        "Package not production ready",
        "The project is marked as non-production ready.",
    )
    INACTIVE_RELEASE = (
        "BASIC006",
        FindingLevel.WARNING,
        "Package is marked as inactive",
        "The project is marked as inactive.",
    )

    def __init__(self, id: str, level: FindingLevel, title: str, message: str):
        """Constructor."""  # noqa: DAR101
        self.id = id
        self.title = title
        self.category = FindingCategory.PROJECT.value
        self.level = level
        self.message = message

    def generate_finding(
        self,
        coordinates: PackageCoordinates,
        message: str = None,
        data: Dictionizer = None,
        url: str = "",
    ) -> Finding:
        """Preps a finding.

        Args:
            coordinates: The package coordinates
            message: Helpful message.
            data: A Dictionizer subclass instance with additional data.
            url: A helpful link for the reader.

        Returns:
            The configured finding.
        """
        if message:
            msg = message
        else:
            msg = self.message

        return Finding(
            coordinates=coordinates,
            id=self.id,
            title=self.title,
            category=self.category,
            level=self.level,
            message=msg,
            data=data,
            url=url,
        )


class BasicReportProvider(BaseReportProvider):
    """Basic report provider implementation.

    Examines the package metadata and calls out any concerns.
    """

    @classmethod
    def get_report_provider_details(cls) -> ReportProviderDetails:
        """Returns the provider details."""
        return ReportProviderDetails(  # noqa: DAR201
            name="basic", vendor="Valiant", display_name="Basic", version="0.1", url=""
        )

    def generate_report(self, package_metadata: PackageMetadata) -> Report:
        """Constructs the report.

        Args:
            package_metadata: containing at least the package metadata

        Returns:
            The report.
        """
        report = Report(BasicReportProvider.get_report_provider_details())

        if not package_metadata.license:
            report.add_finding(
                BasicId.LICENCE_NOT_FOUND.generate_finding(
                    coordinates=package_metadata.coordinates
                )
            )

        if not package_metadata.url_project:
            report.add_finding(
                BasicId.PROJECT_NOT_FOUND.generate_finding(
                    coordinates=package_metadata.coordinates
                )
            )

        if not package_metadata.url_code:
            report.add_finding(
                BasicId.CODEBASE_NOT_FOUND.generate_finding(
                    coordinates=package_metadata.coordinates
                )
            )

        for c in [
            item
            for item in package_metadata.classifiers_parsed
            if item.category == "Development Status"
        ]:
            if c.value in ["7 - Inactive"]:
                report.add_finding(
                    BasicId.INACTIVE_RELEASE.generate_finding(
                        coordinates=package_metadata.coordinates, data=c
                    )
                )
            elif c.value not in ["6 - Mature", "5 - Production/Stable"]:
                report.add_finding(
                    BasicId.NON_PRODUCTION_RELEASE.generate_finding(
                        coordinates=package_metadata.coordinates,
                        data=c,
                        message=f"The package is marked as '{c.value}'",
                    )
                )

        for item in package_metadata.artifacts:
            if not item.signed:
                report.add_finding(
                    BasicId.ARTIFACT_NOT_SIGNED.generate_finding(
                        coordinates=package_metadata.coordinates,
                        message=f"A package of type {item.package_type} has not been signed",
                        data=item,
                    )
                )

        log.info(
            f"Basic reporter located {len(report.findings)} findings"
            " for {str(package_metadata.coordinates)}",
            package_name=package_metadata.name,
            package_version=package_metadata.version,
            repository_url=package_metadata.repository_url,
        )

        return report



"""Report provider relating to SPDX (https://spdx.org/about)."""
from enum import Enum
from typing import Any, Dict, List, Optional

from valiant.log import get_logger
from valiant.package import (
    CLASSIFIER_CATEGORY_LICENSE,
    Classifier,
    PackageCoordinates,
    PackageMetadata,
)
from valiant.reports import (
    BaseReportProvider,
    Finding,
    FindingCategory,
    FindingLevel,
    Report,
    ReportProviderConfiguration,
    ReportProviderDetails,
)
from valiant.util import Dictionizer

from .license import SPDX_LICENSE_DATA_FILE_URL, SpdxLicense, SpdxLicenses


log = get_logger()


class SpdxId(Enum):
    """Setup for findings."""

    FOUND = ("SPDX001", "SPDX License found", FindingLevel.INFO)
    NOT_FOUND = ("SPDX002", "SPDX License not found", FindingLevel.INFO)
    DEPRECATED = ("SPDX003", "Deprecated license", FindingLevel.WARNING)
    NOT_OSI_APPROVED = (
        "SPDX004",
        "License not OSI approved",
        FindingLevel.INFO,
        "https://opensource.org/licenses",
    )
    NOT_FSF_FREE = (
        "SPDX005",
        "License not FSF free software",
        FindingLevel.INFO,
        "https://www.gnu.org/licenses/license-list.html",
    )

    def __init__(  # noqa:D107
        self,
        id: str,
        title: str,
        level: FindingLevel,
        url: str = "https://spdx.org/licenses/",
    ):
        self.id = id
        self.title = title
        self.category = FindingCategory.LICENSE.value
        self.level = level
        self.url = url

    def generate_finding(
        self,
        coordinates: PackageCoordinates,
        message: str,
        data: Dictionizer = None,
        url: str = "",
    ) -> Finding:
        """Preps a finding.

        Args:
            coordinates: The package coordinates
            message: The finding message.
            data: Supplemental data.
            url: A link to help the reader.

        Returns:
            The configured finding.
        """
        return Finding(
            coordinates=coordinates,
            id=self.id,
            title=self.title,
            category=self.category,
            level=self.level,
            message=message,
            data=data,
            url=url if url else self.url,
        )


class SpdxLicenseReportProvider(BaseReportProvider):
    """SPDX License report provider implementation.

    Attempts to match the license provided in package metadata to
    an SPDX license.
    """

    def __init__(self, configuration: ReportProviderConfiguration):  # noqa:D107
        super().__init__(configuration)
        self._licenses = SpdxLicenses.builtin_loader()

    @classmethod
    def get_report_provider_details(cls) -> ReportProviderDetails:
        """Returns the provider details."""
        return ReportProviderDetails(  # noqa: DAR201
            name="spdx",
            vendor="Valiant",
            display_name="SPDX License",
            version="0.1",
            url="https://spdx.org/licenses/",
        )

    @staticmethod
    def _extract_licences_from_metadata(
        package_metadata: PackageMetadata,
    ) -> List[Classifier]:
        candidates: List[Classifier] = []

        if package_metadata.license:
            candidates.append(
                Classifier(
                    original=f"{CLASSIFIER_CATEGORY_LICENSE} :: {package_metadata.license}",
                    category=CLASSIFIER_CATEGORY_LICENSE,
                    subcategories=[],
                    value=package_metadata.license,
                )
            )

        """
        TODO: It might be possible to map the classifiers back to a license code but
              my initial analysis indicated that the classifier info is too coarse

        for item in [
            classifier
            for classifier in package_metadata.classifiers_parsed
            if classifier.category == CLASSIFIER_CATEGORY_LICENSE
        ]:
            candidates.append(item)
        """

        return candidates

    def _map_to_spdx(
        self, package_metadata: PackageMetadata, candidates: List[Classifier]
    ) -> List[Finding]:
        findings: List[Finding] = []

        for item in candidates:
            match = self._licenses.get_license(item.value)
            if match:
                findings.append(
                    SpdxId.FOUND.generate_finding(
                        package_metadata.coordinates,
                        message=match.id,
                        data=match,
                        url=match.details_url,
                    )
                )
                if not match.is_osi_approved:
                    findings.append(
                        SpdxId.NOT_OSI_APPROVED.generate_finding(
                            package_metadata.coordinates,
                            message=f"License '{match.id}' is not OSI-approved.",
                        )
                    )
                if not match.is_fsf_libre:
                    findings.append(
                        SpdxId.NOT_FSF_FREE.generate_finding(
                            package_metadata.coordinates,
                            message=f"License '{match.id}' is not FSF-approved.",
                        )
                    )
            else:
                findings.append(
                    SpdxId.NOT_FOUND.generate_finding(
                        package_metadata.coordinates,
                        message=f"Could not map licence {item.value} to an SPDX license",
                        data=item,
                        url="https://spdx.org/licenses/",
                    )
                )

        return findings

    def map_package_license_to_spdx(
        self, package_metadata: PackageMetadata,
    ) -> List[Finding]:
        """Attempts to map all licence info to SPDX licenses.

        Args:
            package_metadata: The package to review.

        Returns:
            A list of findings. Some will provide a mapping others will not.
        """
        candidates = SpdxLicenseReportProvider._extract_licences_from_metadata(
            package_metadata
        )

        if len(candidates) == 0:
            return []

        return self._map_to_spdx(package_metadata, candidates)

    def generate_report(self, package_metadata: PackageMetadata) -> Report:
        """Constructs the report.

        Args:
            package_metadata: containing at least the package metadata

        Returns:
            The report.
        """
        report = Report(self.get_report_provider_details())
        report.add_findings(self.map_package_license_to_spdx(package_metadata))

        log.info(
            f"SPDX reporter located {len(report.findings)} findings"
            f"for {str(package_metadata.coordinates)}",
            package_name=package_metadata.name,
            package_version=package_metadata.version,
            repository_url=package_metadata.repository_url,
        )

        return report



"""CLI Application config."""
from cleo.config import ApplicationConfig as BaseApplicationConfig  # pragma: no cover
from clikit.api.formatter import Style  # pragma: no cover


class ApplicationConfig(BaseApplicationConfig):  # pragma: no cover
    """Configures the CLI application."""

    _STYLES = [
        Style("c1").fg("cyan"),
        Style("info").fg("blue"),
        Style("comment").fg("green"),
        Style("error").fg("red").bold(),
        Style("warning").fg("yellow"),
        Style("debug").fg("black").bold(),
        Style("link").fg("blue").underlined(),
        Style("h1").fg("blue").bold().underlined(),
        Style("h2").fg("blue").underlined(),
    ]

    def configure(self) -> None:  # noqa: D102
        super(ApplicationConfig, self).configure()

        for style in ApplicationConfig._STYLES:
            self.add_style(style)



"""Base class for Python repos."""

from abc import ABC, abstractmethod
from pathlib import Path
from typing import List

from ..package import PackageMetadata
from .config import RepositoryConfiguration


class BaseRepository(ABC):
    """Interface definition for repo functionality."""

    def __init__(
        self, repository_configuration: RepositoryConfiguration,
    ):
        """New instance.

        The repository_config argument trumps the base_url argument.

        Args:
            repository_configuration: A RepositoryConfiguration instance

        Raises:
            ValueError: If no configuration is provided or
                        no handler is available for the repo type.
        """
        if (
            repository_configuration.repository_type
            not in self.__class__.list_supported_repository_types()
        ):
            raise ValueError(
                f"Unable to handle repositories of type "
                "repository_configuration.repository_type}."
            )

        self._config = repository_configuration

    @property
    def repository_configuration(self) -> RepositoryConfiguration:
        """Get the config."""
        return self._config  # noqa: DAR201

    @property
    def repository_type(self) -> str:
        """Identifies the underlying software supported by the repo.

        Returns:
            The repo type.
        """
        return self._config.repository_type

    @classmethod
    def list_supported_repository_types(cls) -> List[str]:
        """Lists the repository types support by this implementation."""
        return []  # noqa: DAR201

    @abstractmethod
    def show(self, name: str, version: str) -> PackageMetadata:
        """Provides details for a specific package version.

        Args:
            name: The package name
            version: The package version

        # noqa: DAR202
        Returns:
            A package instance if it can be located.

        Raises:
            NotImplementedError: Because this is an abstract implementation.
            PackageNotFoundException: When the package cannot be found. # noqa: DAR402
            ValidationError: The package metadata could not be parsed. # noqa: DAR402
        """
        raise NotImplementedError

    @abstractmethod
    def download(self, name: str, version: str) -> Path:
        """Provides details for a specific package version.

        Args:
            name: The package name
            version: The package version

        # noqa: DAR202
        Returns:
            The path to the download if the artifact can be located.

        Raises:
            NotImplementedError: Because this is an abstract implementation.
        """
        raise NotImplementedError



"""Integrating the Safety library.

See: https://github.com/pyupio/safety
"""
from enum import Enum
from typing import Dict, List, Optional
from safety import __author__, __version__
from safety.safety import check as safety_check, Vulnerability
from safety.util import Package as SafetyPackage

from ...package import PackageCoordinates, PackageMetadata
from .. import (
    BaseReportProvider,
    Report,
    ReportProviderDetails,
    FindingLevel,
    FindingCategory,
    Finding,
)
from valiant.log import get_logger
from valiant.util import Dictionizer

log = get_logger()


class VulnerabilityDictionizer(Dictionizer):
    """A wrapper for Vulnerability to allow dict and json transforms."""

    def __init__(self, vulnerability: Vulnerability):
        """Constructor."""  # noqa:DAR101
        self._vulnerability = vulnerability

    def to_dict(self) -> Dict:  # noqa:DAR201
        return self._vulnerability._asdict()


class SafetyId(Enum):
    """Setup for findings."""

    VULNERABILITY_FOUND = "SAFETY001"

    def __init__(self, id: str):
        """Constructor."""  # noqa: DAR101
        self.id = id
        self.category = FindingCategory.SECURITY.value
        self.level = FindingLevel.PRIORITY

    def generate_finding(
        self, coordinates: PackageCoordinates, vulnerability: Vulnerability
    ) -> Finding:
        """Preps a finding.

        Args:
            coordinates: The package coordinates
            vulnerability: A vulnerability found by Safety

        Returns:
            The configured finding.
        """
        return Finding(
            coordinates=coordinates,
            id=self.id,
            category=self.category,
            level=self.level,
            title="Vulnerability found",
            message=vulnerability.advisory,
            data=VulnerabilityDictionizer(vulnerability),
            url="https://github.com/pyupio/safety-db",
        )


class SafetyReportProvider(BaseReportProvider):
    """Provider implementation."""

    @classmethod
    def get_report_provider_details(cls) -> ReportProviderDetails:
        """Returns the provider details."""
        return ReportProviderDetails(  # noqa: DAR201
            name="safety",
            vendor=__author__,
            display_name="Safety",
            version=__version__,
            url="https://pyup.io/safety/",
        )

    def generate_report(self, package_metadata: PackageMetadata) -> Report:
        """Constructs the report.

        Args:
            package_metadata: containing at least the package metadata

        Returns:
            The report.
        """
        # Prep config
        if self.configuration:
            key: Optional[str] = self.configuration.get("key", None)
            db: Optional[str] = self.configuration.get("db", None)
            ignore_ids: Optional[str] = self.configuration.get("ignore_ids", None)
        else:
            key = None
            db = None
            ignore_ids = None

        if ignore_ids:
            ignore_id_list = [id.strip() for id in ignore_ids.split(",")]
        else:
            ignore_id_list = []

        packages = [
            SafetyPackage(key=package_metadata.name, version=package_metadata.version)
        ]

        vulnerabilities: List[Vulnerability] = safety_check(
            packages=packages,
            key=key,
            db_mirror=db,
            cached=False,
            ignore_ids=ignore_id_list,
            proxy=None,
        )

        report = Report(SafetyReportProvider.get_report_provider_details())

        for v in vulnerabilities:
            report.add_finding(
                SafetyId.VULNERABILITY_FOUND.generate_finding(
                    package_metadata.coordinates, v
                )
            )

        log.info(
            f"Safety reporter located {len(report.findings)} findings"
            f" for {str(package_metadata.coordinates)}",
            package_name=package_metadata.name,
            package_version=package_metadata.version,
            repository_url=package_metadata.repository_url,
        )

        return report



"""CLI Command: about."""
import json
from typing import Optional
from .base_command import BaseCommand


class AboutCommand(BaseCommand):
    """Shows information about Valiant.

    about
        {--o|out= : the desired output type (json)}
    """

    def handle(self) -> Optional[int]:  # noqa: D102
        format = self.option("out")
        if format == "json":
            self.line(self.to_json())
        else:
            self.line(self.to_text())
        return 0

    def to_json(self) -> str:  # noqa: D102
        return json.dumps(
            {
                "application": {
                    "name": self.valiant.application_name,
                    "version": self.valiant.application_version,
                    "license": self.valiant.application_licence,
                    "url": self.valiant.application_homepage,
                }
            }
        )

    def to_text(self) -> str:  # noqa: D102
        return f"""\
<info>{self.valiant.application_title} {self.valiant.application_version}</info>
<comment>- {self.valiant.application_tagline} - </comment>

<comment>{self.valiant.application_description}</comment>

<comment>Licence: {self.valiant.application_licence}</comment>
<comment>See <link>{self.valiant.application_homepage}</link> for more information.</comment>"""



"""Configuration for describing Python repositories."""
from dataclasses import asdict, dataclass
from typing import Dict, Optional
from urllib.parse import ParseResult, urlparse, urlunparse

from valiant.util import Dictionizer


@dataclass(frozen=True)
class RepositoryConfiguration(Dictionizer):
    """A Python package repository.

    Working from https://pip.pypa.io/en/stable/user_guide/#basic-authentication-credentials

    Attributes:
        name: A human-friendly name
        base_url: The repository URL such as https://pypi.org/pypi
                  Don't include things like `/simple` in the base_url.
        username: A username for the password
        password: The access password (accompanies a username)
        token: An access token (used instead of username:password)
        repository_type: The software used to host the repository
    """

    name: str
    base_url: str
    username: Optional[str] = None
    password: Optional[str] = None
    token: Optional[str] = None
    repository_type: str = "unknown"

    def to_dict(self) -> Dict:  # noqa:D102
        return asdict(self)

    def get_access_url(self) -> str:
        """Provides the full url with credentials.

        Returns:
            The access url for the repository
        """
        url_segments: ParseResult = urlparse(self.base_url)

        if self.token:
            netloc = f"{self.token}@{url_segments.netloc}"
        elif self.username:
            netloc = f"{self.username}:{self.password}@{url_segments.netloc}"
        else:
            netloc = url_segments.netloc

        url = ParseResult(
            scheme=url_segments.scheme,
            netloc=netloc,
            path=url_segments.path,
            params=url_segments.params,
            query=url_segments.query,
            fragment=url_segments.fragment,
        )
        return urlunparse(url)



"""Base class for report providers."""
from abc import ABC, abstractmethod
from typing import Any, Dict, Optional

from ..package import PackageMetadata
from .model import Report, ReportProviderConfiguration, ReportProviderDetails


class BaseReportProvider(ABC):
    """Base class for report providers."""

    def __init__(self, configuration: ReportProviderConfiguration):
        """Constructor.

        Args:
            configuration: The provider's config
        """
        self._configuration = configuration

    @property
    def configuration(self) -> Optional[Dict[str, Any]]:
        """The provider config."""
        return self._configuration.items  # noqa: DAR201

    @classmethod
    @abstractmethod
    def get_report_provider_details(cls) -> ReportProviderDetails:  # pragma: no cover
        """Returns the provider details."""
        pass

    @abstractmethod
    def generate_report(
        self, package_metadata: PackageMetadata
    ) -> Report:  # pragma: no cover
        """Constructs the report.

        Args:
            package_metadata: containing at least the package metadata

        # noqa: DAR202
        Returns:
            The report.
        """
        pass



"""Valiant CLI client."""
from cleo import Application as BaseApplication
from valiant import Factory, Valiant

from .commands import (
    AboutCommand,
    AuditCommand,
    ConfigCommand,
    ReportCommand,
    ShowCommand,
)
from .config import ApplicationConfig


class Cli(BaseApplication):
    """The CLI client app."""

    def __init__(self):
        """Constructor."""
        self._valiant: Valiant = Factory().create_valiant()
        super(Cli, self).__init__(
            name=self._valiant.application_name,
            version=self._valiant.application_version,
            config=ApplicationConfig(
                self._valiant.application_name, self._valiant.application_version,
            ),
        )
        self.add_commands(
            AboutCommand(),
            ConfigCommand(),
            AuditCommand(),
            ReportCommand(),
            ShowCommand(),
        )

    @property
    def valiant(self) -> Valiant:
        """Access the Valiant instance."""
        return self._valiant  # noqa: DAR201


if __name__ == "__main__":
    Cli().run()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['valiant',
 'valiant.config',
 'valiant.console',
 'valiant.console.commands',
 'valiant.console.config',
 'valiant.log',
 'valiant.package',
 'valiant.reports',
 'valiant.reports.basic',
 'valiant.reports.github',
 'valiant.reports.librariesio',
 'valiant.reports.safety',
 'valiant.reports.spdx',
 'valiant.repositories',
 'valiant.repositories.pypi',
 'valiant.util']

package_data = \
{'': ['*']}

install_requires = \
['appdirs>=1.4.3,<2.0.0',
 'cleo>=0.7.6,<0.8.0',
 'desert>=2020.1.6,<2021.0.0',
 'marshmallow>=3.5.1,<4.0.0',
 'requests-cache>=0.5.2,<0.6.0',
 'requests>=2.23.0,<3.0.0',
 'safety>=1.8.7,<2.0.0',
 'setuptools>=46.0.0,<47.0.0',
 'structlog>=20.1.0,<21.0.0',
 'texttable>=1.6.2,<2.0.0',
 'wcwidth>=0.1.8,<0.2.0']

entry_points = \
{'console_scripts': ['valiant = valiant.console:main']}

setup_kwargs = {
    'name': 'valiant',
    'version': '0.1.0',
    'description': 'Audit tool to help investigate Python dependencies',
    'long_description': '# Valiant\n[![pre-commit](https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white)](https://github.com/pre-commit/pre-commit)\n\nThe Valiant project aims to provide auditing tools that help project teams\ntrack their dependencies in terms of licensing, security, and dependability.\n\nThe goal is to help reduce the fragility of the "input side" of software development\nby making it easy to assess and track dependencies. Further work will also aim to\ncheck dependencies against policy documents to help\n\nThe system is written in Python 3 and targets Python codebases.\n\nPlease check out the [project site](http://pomes.github.io/valiant)\nfor documentation.\n\n## Key resources\n\n| Resource | Description |\n| -------- | ----------- |\n| [Codebase](https://github.com/pomes/valiant) | GitHub project |\n| [Security policy](SECURITY.md) | Please refer to this if you need to report a security concern |\n| [Project tracker](https://github.com/pomes/valiant/projects/1) | Structured to provide fulfilment of [project milestones](https://github.com/pomes/valiant/milestones) |\n| [Issue tracker](https://github.com/pomes/valiant/issues) | GitHub issues |\n| [Documentation](http://pomes.github.io/valiant) | Project site |\n| [License](LICENSE.txt) | Project licence (MIT) |\n| [Contributing](CONTRIBUTING.md) | Want to join in? |\n\n## Attribution\n\nI have used the [Poetry](https://github.com/python-poetry/poetry) library to package\nthis project and guide my implementation efforts.\n\nThe [Hypermodern Python series](https://cjolowicz.github.io/posts/hypermodern-python-01-setup/) by\nClaudio Jolowicz is a great resource and you\'ll see his guidance instilled in this codebase.\n',
    'author': 'Duncan Dickinson',
    'author_email': 'dedickinson@users.noreply.github.com',
    'maintainer': 'Duncan Dickinson',
    'maintainer_email': 'dedickinson@users.noreply.github.com',
    'url': 'https://github.com/pomes/valiant',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""Data model for the PyPi JSON API.

The data classes defined here aim to map closely to the JSON
returned by the PyPi API.

See: https://warehouse.readthedocs.io/api-reference/json/

TODO: Consider "yanked" artifacts: https://www.python.org/dev/peps/pep-0592/
"""

from dataclasses import dataclass
from datetime import datetime
from typing import Dict, List, Optional

from desert import desert, marshmallow
from packaging.requirements import Requirement
from valiant.package import (
    ArtifactMetadata,
    ArtifactMetadataImpl,
    Classifier,
    PackageMetadata,
)

from .. import ValidationError


@dataclass
class Downloads:
    """The (unused?) download counts for a package.

    See https://packaging.python.org/guides/analyzing-pypi-package-downloads/
    """

    last_day: int
    last_month: int
    last_week: int


@dataclass
class Info:
    """The package metadata.

    See also: https://packaging.python.org/specifications/core-metadata/
    """

    author: str
    author_email: str
    bugtrack_url: Optional[str]
    classifiers: List[str]
    description: str
    description_content_type: Optional[str]
    docs_url: Optional[str]
    download_url: str
    downloads: Downloads
    home_page: str
    keywords: Optional[str]
    license: str
    maintainer: Optional[str]
    maintainer_email: Optional[str]
    name: str
    package_url: str
    platform: str
    project_url: str
    project_urls: Dict[str, str]
    release_url: str
    requires_dist: Optional[List[str]]
    requires_python: Optional[str]
    summary: str
    version: str


@dataclass
class Release:
    """A single release artifact for a package.

    PyPi seems to return a list of all releases regardless
    of if you designate a version.
    """

    comment_text: str
    digests: Dict[str, str]
    downloads: int
    filename: str
    has_sig: bool
    md5_digest: str
    packagetype: str
    python_version: str
    requires_python: Optional[str]
    size: int
    upload_time: datetime
    upload_time_iso_8601: datetime
    url: str


@dataclass
class ArtifactUrl:
    """An artifact (i.e. file) for a package."""

    comment_text: str
    digests: Dict[str, str]
    downloads: int
    filename: str
    has_sig: bool
    md5_digest: str
    packagetype: str
    python_version: str
    requires_python: Optional[str]
    size: int
    upload_time: datetime
    upload_time_iso_8601: datetime
    url: str


@dataclass
class PyPiPackage:
    """Represents the package metadata returned from pypi.org."""

    info: Info
    last_serial: int
    releases: Dict[str, List[Release]]
    urls: List[ArtifactUrl]


@dataclass
class PyPiPackageMetadata(PackageMetadata):
    """Provides the required PackageMetadata interface for a PyPiPackage."""

    def __init__(self, repository_url: str, package_data: Dict):
        """Constructor.

        Args:
            repository_url: The URL for the repo that provided this metadata.
            package_data: A dictionary based on the JSON structure returned by the PyPi API.
                          See: https://warehouse.readthedocs.io/api-reference/json/

        Raises:
            ValidationError: when the data could not be correctly mapped.
        """
        self._repository_url = repository_url
        self._parsed_classifiers: Optional[List[Classifier]] = None
        self._artifacts: List[ArtifactMetadata] = []

        try:
            self._pkg = desert.schema(PyPiPackage).load(package_data)
        except marshmallow.exceptions.ValidationError as ve:
            raise ValidationError(f"Could not validate the JSON data: {ve}") from ve

        self._requires_dist: Dict[str, List[Requirement]] = {}
        if self._pkg.info.requires_dist:
            for item in self._pkg.info.requires_dist:
                req = Requirement(item)
                if req.name not in self._requires_dist:
                    self._requires_dist[req.name] = [req]
                else:
                    self._requires_dist[req.name].append(req)

        # Process the urls into artifacts
        for entry in self._pkg.urls:
            self._artifacts.append(
                ArtifactMetadataImpl(
                    comment_text=entry.comment_text,
                    digests=entry.digests,
                    sha256_digest=entry.digests.get("sha256", None),
                    signed=entry.has_sig,
                    signature_url=f"{entry.url}.asc",
                    package_type=entry.packagetype,
                    python_version=entry.python_version,
                    requires_python=entry.requires_python,
                    size=entry.size,
                    upload_time_iso_8601=entry.upload_time_iso_8601,
                    url=entry.url,
                )
            )

    @property
    def name(self) -> str:  # noqa: D102
        return self._pkg.info.name

    @property
    def description(self) -> str:  # noqa: D102
        return self._pkg.info.description

    @property
    def summary(self) -> str:  # noqa: D102
        return self._pkg.info.summary

    @property
    def license(self) -> str:  # noqa: D102
        return self._pkg.info.license

    @property
    def classifiers(self) -> List[str]:  # noqa: D102
        """The classifiers listed against the package.

        Returns:
            A list of classifier strings.
        """
        return self._pkg.info.classifiers

    @property
    def classifiers_parsed(self) -> List[Classifier]:  # noqa: D102
        """The parsed classifiers listed against the package.

        Returns:
            A list of parsed classifiers.
        """
        if not self._parsed_classifiers:
            self._parsed_classifiers = []
            for item in self.classifiers:
                self._parsed_classifiers.append(Classifier.parse(item))

        return self._parsed_classifiers

    @property
    def version(self) -> str:  # noqa: D102
        return self._pkg.info.version

    @property
    def repository_url(self) -> str:  # noqa: D102
        return self._repository_url

    @property
    def url_code(self) -> str:  # noqa: D102
        return self._pkg.info.project_urls.get("Code", "")

    @property
    def url_documentation(self) -> str:  # noqa: D102
        if self._pkg.info.docs_url:
            return self._pkg.info.docs_url

        return self._pkg.info.project_urls.get("Documentation", "")

    @property
    def url_project(self) -> str:  # noqa: D102
        if self._pkg.info.home_page:
            return self._pkg.info.home_page

        if self._pkg.info.project_url:
            return self._pkg.info.project_url

        return self._pkg.info.project_urls.get("Homepage", "")

    @property
    def url_issue_tracker(self) -> str:  # noqa: D102
        if self._pkg.info.bugtrack_url:
            return self._pkg.info.bugtrack_url

        return self._pkg.info.project_urls.get("Issue tracker", "")

    @property
    def requires_dist(self) -> Dict[str, List[Requirement]]:
        """A distutils-based dependency.

        See:
        https://packaging.python.org/specifications/core-metadata/#requires-dist-multiple-use

        See: PEP 508 -- Dependency specification for Python Software Packages
        https://www.python.org/dev/peps/pep-0508/

        See:
        https://setuptools.readthedocs.io/en/latest/pkg_resources.html#requirement-methods-and-attributes # noqa: B950

        Versioning spec:
            https://www.python.org/dev/peps/pep-0440/#version-specifiers

        Returns:
            A dictionary of parsed Requirements with the requirement name as key
            See: (https://packaging.pypa.io/en/latest/requirements/)
        """
        return self._requires_dist

    @property
    def requires_python(self) -> List[str]:
        """Lists the Python version specifiers.

        Packaging spec:
        https://packaging.python.org/specifications/core-metadata/#requires-python

        Versioning spec:
        https://www.python.org/dev/peps/pep-0440/#version-specifiers

        Returns:
            The list of requires-python entries
        """
        return [p.strip() for p in self._pkg.info.requires_python.split(",")]

    @property
    def artifacts(self) -> List[ArtifactMetadata]:
        """A distutils-based dependency.

        See:
        https://packaging.python.org/specifications/core-metadata/#requires-dist-multiple-use

        See: PEP 508 -- Dependency specification for Python Software Packages
        https://www.python.org/dev/peps/pep-0508/


        See:
        https://setuptools.readthedocs.io/en/latest/pkg_resources.html#requirement-methods-and-attributes # noqa: B950

        Versioning spec:
            https://www.python.org/dev/peps/pep-0440/#version-specifiers


        Returns:
            A dictionary of parsed Requirements with the requirement name as key
        """
        return self._artifacts



"""Base command for use by commands that use a basic payload."""
from abc import abstractmethod
from dataclasses import dataclass
from typing import Dict, Optional

from valiant.package import PackageMetadata
from valiant.reports import ReportSet
from valiant.util import Dictionizer

from .base_command import BaseCommand


@dataclass(frozen=True)
class Payload(Dictionizer):
    """Utility class for passing data around."""

    metadata: Optional[PackageMetadata] = None
    reports: Optional[ReportSet] = None
    message: str = ""

    def to_dict(self) -> Dict:  # noqa:D102
        return {
            "message": self.message,
            "metadata": self.metadata.to_dict() if self.metadata else None,
            "reports": self.reports.to_dict() if self.reports else None,
        }


class PackageCommand(BaseCommand):
    """Handy general command base for use by actual commands."""

    def handle(self) -> Optional[int]:  # noqa: D102
        try:
            data = self.prepare_data()
            self.output_data(data, format=self.option("out"))
            return 0
        except Exception as e:
            self.output_error(e, format=self.option("out"))
            return 1

    @abstractmethod
    def prepare_data(self) -> Payload:
        """Subclass-based payload prep."""
        raise NotImplementedError  # noqa:DAR101,DAR401

    def output_data(self, payload: Payload, format: str = None) -> None:
        """Outputs the payload to the required format."""  # noqa:DAR101
        if format == "json":
            self.line(self.to_json(payload))
        else:
            self.line(self.to_text(payload))

    def to_text(self, payload: Payload) -> str:
        """Prepares text representations.

        Args:
            payload: The payload to display.

        Returns:
            At the very least we'll return `payload.message`.
        """
        return payload.message

    def to_json(self, payload: Payload) -> str:
        """Converts payload to json."""  # noqa:DAR101,DAR201
        import json

        if self.option("short"):
            if payload.reports:
                return json.dumps([f.to_dict() for f in payload.reports.all_findings])
            else:
                return "[]"  # empty json list

        return payload.to_json()
